import Stepper from 'bs-stepper';
import 'select2/dist/js/select2.full.min';
import 'bootstrap-switch';

function updateIterators(disableFirst: boolean): void {
    // Get all iterators
    const iterators: HTMLCollectionOf<Element> = document.getElementsByClassName("iterator");

    // For each iterator get iterator class and use dictionary
    let dict = new Map<string, Element[]>();
    for (let i: number = 0; i < iterators.length; i++) {
        const elem = iterators.item(i);
        if (!elem) { continue; }
        for (let j: number = 0; j < elem.classList.length; j++) {
            const className: string = elem.classList[j];
            if (!className) { continue; }
            if (className.endsWith("-iterator")) {
                let pastValue = dict.get(className)
                if (!pastValue) pastValue = [];
                pastValue.push(elem);
                dict.set(className, pastValue);
                continue;
            }
        }
    }

    dict.forEach((elems: Element[]) => {
        for (let i: number = 0; i < elems.length; i++) {
            const elem = elems[i] as HTMLSpanElement;

            const parentLi = elem.closest("li") as HTMLLIElement;
            const deleteButton = parentLi.querySelector(".remove_item_link") as HTMLButtonElement;
            deleteButton.disabled = disableFirst && (elems.length == 1);

            elem.innerText = (i + 1).toString();
        }
    });
}

function fixFileInputLabel(item: HTMLLIElement) {
    (item.querySelectorAll('input') as NodeListOf<HTMLInputElement>).forEach((input) => {
        if (input.type == "file") {
            input.labels?.forEach((label: HTMLLabelElement) => {
                label.textContent = "Choisissez un fichier...";
            });
        }

        input.addEventListener('input', (event: InputEvent) => {
            input.classList.toggle('is-invalid', !input.checkValidity());

            // Update label on file upload
            if (input.type === "file") {
                input.labels?.forEach((label: HTMLLabelElement) => {
                    if (input.files && input.files[0]) {
                        label.textContent = input.files[0].name;
                    }
                });
            }
        });
    });
}

function addFormToCollection(e: any): void {
    const collectionHolder = document.querySelector('.' + e.currentTarget.dataset.collectionHolderClass) as any;
    const ul = document.querySelector('ul.' + e.currentTarget.dataset.collectionHolderClass);
    const disableFirst = ul?.getAttribute('required') === '1'; // undefined is set false

    const item = document.createElement('li');

    item.innerHTML = collectionHolder
        .dataset
        .prototype
        .replace(
            /__name__/g,
            collectionHolder.dataset.index
        );

    fixFileInputLabel(item);
    collectionHolder.appendChild(item);

    collectionHolder.dataset.index++;
    configureDeleteButton(item, disableFirst);
    updateIterators(disableFirst);
}

function configureDeleteButton(item: HTMLLIElement, disableFirst: boolean): void {
    // Setup Select inputs
    ($('select', item) as any).select2({
        language: 'fr',
        theme: 'bootstrap4',
        minimumResultsForSearch: Infinity
    });
    const deleteButton = item.querySelector(".remove_item_link") as HTMLButtonElement;
    const iterator = item.querySelector(".iterator") as HTMLSpanElement;

    // Disable first if fiels is required
    if (deleteButton) {
        if (disableFirst && iterator.innerText == "1") {
            deleteButton.disabled = true;
        }

        deleteButton.addEventListener('click', (e) => {
            e.preventDefault();
            item.remove();
            updateIterators(disableFirst);
        });
    }
}

function configureDefaultSubForm(addButton: HTMLButtonElement): void {
    const items = document.querySelectorAll('.' + addButton.dataset.collectionHolderClass + ' li') as NodeListOf<HTMLLIElement>;
    const ul = document.querySelector('ul.' + addButton.dataset.collectionHolderClass);
    const disableFirst = ul?.getAttribute('required') === '1'; // undefined is set false

    items.forEach(item => {
        configureDeleteButton(item, disableFirst);
        fixFileInputLabel(item);
    });

    // Add first element if required
    if (disableFirst && items.length == 0) {
        addFormToCollection({ currentTarget: addButton });
    }
}

function getStepperElement(): HTMLDivElement {
    const stepperElem = document.querySelector('.bs-stepper') as HTMLDivElement;
    if (stepperElem == null) {
        throw "Could not find stepper Element"
    }
    return stepperElem;
}

function getStepperContentTabElement(tabId: number): HTMLDivElement {
    const stepperElem = getStepperElement();
    const stepperTab = stepperElem.querySelectorAll(".content")[tabId] as HTMLDivElement;
    if (stepperTab == null) {
        throw "Could not find stepper Tab " + tabId;
    }
    return stepperTab;
}

function getStepperTabButton(tabId: number): HTMLButtonElement {
    const stepperTab = getStepperContentTabElement(tabId);
    const tabIdentifier = stepperTab.getAttribute('aria-labelledby');
    if (tabIdentifier == null) {
        throw "Could not find tab button label for tab " + tabId;
    }

    const tabBtn = document.getElementById(tabIdentifier) as HTMLButtonElement;

    if (tabBtn == null) {
        throw "Could not find tab button for tab " + tabId;
    }

    return tabBtn;
}

function clearTabErrorBadge(tabId: number) {
    const tabButton = getStepperTabButton(tabId);
    tabButton.querySelectorAll("span.badge").forEach((badge: HTMLDivElement) => {
        tabButton.removeChild(badge);
    });
}

function IncrTabErrorBadge(tabId: number) {
    const tabButton = getStepperTabButton(tabId);
    const currentBadge = tabButton.querySelector("span.badge") as HTMLSpanElement;
    if (currentBadge != null) {
        const currentValue = currentBadge.innerText;
        currentBadge.innerText = (parseInt(currentValue) + 1).toString();
    } else {
        // Create badge and add one error
        const badge = document.createElement('span');
        badge.classList.add('badge', 'bg-danger');
        badge.innerText = '1';
        const icon = tabButton.querySelector('.bs-stepper-circle');
        if (icon) {
            tabButton.insertBefore(badge, icon?.nextSibling);
        }
    }
}

function getTabIdFromChildElem(elem: Element): number {
    const tabContent = elem.closest(".content") as HTMLDivElement;
    if (tabContent == null) {
        throw "Could not find tab for error " + elem;
    }

    let tab = -1;
    const children = tabContent.parentNode?.children;
    if (children == null) {
        throw "Could not find content children";
    }

    for (let i = 0; i < children.length; i++) {
        const child = children.item(i);
        if (tabContent.isSameNode(child)) {
            tab = i;
        }
    }

    if (tab == -1) {
        throw "Could not find Tab Id from elem";
    }
    return tab;
}

function getNumberTabs(): number {
    return document.querySelectorAll('.bs-stepper-content>.content').length;
}

function findAndAddErrorbadges() {
    const nbTabs = getNumberTabs()
    for (let i = 0; i < nbTabs; i++) {
        clearTabErrorBadge(i);
    }
    document.querySelectorAll('.invalid-feedback').forEach((invalidTag: HTMLDivElement) => {
        IncrTabErrorBadge(getTabIdFromChildElem(invalidTag));
    });
}

function CustomAutoAdd(tabId: number, value: boolean) {
    const tab = getStepperContentTabElement(tabId);
    const ul = tab.querySelector('ul');
    if (ul == null) {
        throw "Ul not found";
    }
    if (value) {
        // Get add button
        if (ul.querySelectorAll('li').length == 0) {
            const addButton = tab.querySelector(".add_item_link");
            if (addButton) {
                // Add a required default form
                addFormToCollection({ currentTarget: addButton });
            }
        }
    } else {
        // Delete all elements
        ul.querySelectorAll('li').forEach((item: ChildNode) => ul.removeChild(item));
        clearTabErrorBadge(tabId);
    }
}

$(function () {
    let stepper: Stepper;

    // Setup The form tabs
    const stepperElem = $('.bs-stepper')[0];
    stepper = new Stepper(stepperElem, { linear: false, animation: true });

    const form = document.querySelector('form') as HTMLFormElement;

    const nextButtons: NodeListOf<Element> = document.querySelectorAll('.request-form-next');
    nextButtons.forEach((element: HTMLButtonElement) => {
        element.addEventListener('click', () => {
            stepper.next();
        });
    });

    const prevButtons: NodeListOf<Element> = document.querySelectorAll('.request-form-prev');
    prevButtons.forEach((element: HTMLButtonElement) => {
        element.addEventListener('click', () => {
            stepper.previous();
        });
    });
    // Setup Select inputs
    ($('select') as any).select2({
        language: 'fr',
        theme: 'bootstrap4',
        minimumResultsForSearch: Infinity
    });


    // Adding items to collections
    const buttons = document
        .querySelectorAll('.add_item_link') as NodeListOf<HTMLButtonElement>;

    buttons.forEach(btn => btn.addEventListener("click", addFormToCollection));
    buttons.forEach(btn => configureDefaultSubForm(btn));

    const is_parent_ref: HTMLInputElement | null = document.getElementById('parking_application_parent') as HTMLInputElement;
    const is_parent_tab_trigger: HTMLButtonElement = document.getElementById('parent-info-trigger') as HTMLButtonElement;

    const has_physical_cond_ref: HTMLInputElement | null = document.getElementById('parking_application_physicalCondition') as HTMLInputElement;
    const has_physical_cond_tab_trigger: HTMLButtonElement = document.getElementById('physical-info-trigger') as HTMLButtonElement;
    is_parent_tab_trigger.disabled = !is_parent_ref.checked;
    has_physical_cond_tab_trigger.disabled = !has_physical_cond_ref.checked;
    CustomAutoAdd(2, is_parent_ref.checked);
    CustomAutoAdd(3, has_physical_cond_ref.checked);

    is_parent_ref.addEventListener('input', () => {
        is_parent_tab_trigger.disabled = !is_parent_ref.checked;
        CustomAutoAdd(2, is_parent_ref.checked);
    });
    has_physical_cond_ref.addEventListener('input', () => {
        has_physical_cond_tab_trigger.disabled = !has_physical_cond_ref.checked;
        CustomAutoAdd(3, has_physical_cond_ref.checked);
    });

    stepperElem.addEventListener('show.bs-stepper', (event) => {
        const tab_from: number = (event as any).detail.from;
        const tab_to_get: number = (event as any).detail.to;
        const is_increase = (tab_to_get - tab_from) >= 0;

        if (is_increase) {
            let tab_is_valid = true;
            for (let i = 0; i < tab_to_get; i++) {
                clearTabErrorBadge(i);
                const tab = getStepperContentTabElement(i);

                tab.querySelectorAll('input,textarea').forEach((value: HTMLInputElement) => {
                    if (!value.checkValidity()) {
                        tab_is_valid = false;
                        IncrTabErrorBadge(i);
                    }
                });
            }

            if (!tab_is_valid) {
                form.classList.add("was-validated");
                event.preventDefault();
                return;
            }
        }

        const btn_tab_to_get = getStepperTabButton(tab_to_get);
        if (btn_tab_to_get.disabled) {
            event.preventDefault();
            stepper.to(tab_to_get + (is_increase ? 2 : 0));
        }
    });

    stepperElem.addEventListener('shown.bs-stepper', (event) => {
        form.classList.remove("was-validated");
    });

    (document
        .querySelector('#parking_application_carpooling') as HTMLInputElement).onchange = ((event: InputEvent) => {
            // Disable description
            const descriptionElem = document.querySelector('#parking_application_carpoolDescription') as HTMLInputElement;
            const carpoolValue = (event.target as HTMLInputElement).value == '1';
            descriptionElem.disabled = !carpoolValue;
            descriptionElem.required = carpoolValue;
        });

    $("input[data-bootstrap-switch]").each(function () {
        const elem = $(this)[0] as HTMLInputElement;
        ($(this) as any).bootstrapSwitch('state', elem.value == '1');
        ($(this) as any).bootstrapSwitch('onSwitchChange', (event: Event, state: boolean) => {
            elem.value = state ? '1' : '0';
            elem.dispatchEvent(new InputEvent("change", { target: elem } as any));
        });
    });
    findAndAddErrorbadges();
});
