<?php

namespace App\Controller\Admin;

use App\Entity\File;
use App\Entity\ParkingApplication;
use App\Enum\ApplicationStatusEnum;
use App\Enum\EmailEnum;
use App\Form\ParkingApplicationType;
use App\Repository\ParkingApplicationRepository;
use App\Repository\SessionRepository;
use App\Service\ApplicationService;
use App\Service\MailerService;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\Exception\BadRequestException;
use Symfony\Component\HttpFoundation\File\Exception\FileNotFoundException;
use Symfony\Component\HttpFoundation\File\File as HTTPFile;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\ResponseHeaderBag;
use Symfony\Component\Mime\Address;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Validator\Validator\ValidatorInterface;
use Symfony\Contracts\Translation\TranslatorInterface;

#[Route('/admin/application', name: 'admin_application')]
class ApplicationController extends AbstractController
{
    # Default get of all applications, sorted by time
    #[Route('/session/{id}', name: '', methods: ['GET'], defaults: ['id' => null])]
    public function application(
        ParkingApplicationRepository $applicationRepo,
        SessionRepository $sessionRepository,
        TranslatorInterface $translator,
        ?string $id = null,
    ): Response
    {
        $session = $id == null ? $sessionRepository->findActive() : $sessionRepository->find($id);

        if ($session == null) {
            $this->addFlash("error", $translator->trans("flash.no-active-session"));
            return $this->redirectToRoute('admin_config_session');
        }

        $data = $applicationRepo->findBySession($session);
        $approved = array_filter($data, fn($app) => $app->getStatus() == ApplicationStatusEnum::APPROVED);
        $counts = $applicationRepo->countByStatus($session);

        return $this->render(
            'admin/application/index.html.twig',
            [
                "session" => $session,
                "applications" => $data,
                "counts" => $counts,
                "approved_applications" => $approved,
            ]
        );
    }

    #[Route('/{id}/status', name: '_change_status', methods: ['PUT'])]
    public function changeStatus(
        ParkingApplication $parkingApp,
        ValidatorInterface $validator,
        EntityManagerInterface $manager,
        MailerService $mailer,
        Request $request
    ): Response {
        $params = json_decode($request->getContent(), true);

        if (!is_array($params)) {
            throw new BadRequestException("Invalid JSON !");
        }
        if (!array_key_exists("status", $params)) {
            throw new BadRequestException("status not in JSON !");
        }

        $newStatus = ApplicationStatusEnum::tryFrom($params['status']);
        if ($newStatus == null) {
            throw new BadRequestException("Incorrect status !");
        }

        $parkingApp->setStatus($newStatus);
        if (count($validator->validate($parkingApp)) != 0)
            throw new BadRequestException("Parking Application is invalid !");

        $mailer->sendEmail(
            new Address($parkingApp->getEmail()),
            [
                "application" => $parkingApp
            ],
            EmailEnum::APPLICATION_PROCESSED_EMAIL
        );

        $manager->persist($parkingApp);
        $manager->flush();
        return new Response();
    }

    #[Route('/{id}', name: '_get', methods: ['GET'])]
    public function getApplication(ParkingApplication $parkingApp): Response
    {
        return $this->render('admin/application/get.html.twig', [
            'application' => $parkingApp
        ]);
    }

    #[Route('/file/{id}', name: '_get_file', methods: ['GET'])]
    public function getFile(File $file): Response
    {
        $dirname = $this->getParameter('uploads_dir');
        try {
            $fileObj = new HTTPFile($dirname . $file->getFileName());
        } catch (FileNotFoundException $e) {
            throw $this->createNotFoundException("File not found");
        }
        return $this->file($fileObj, disposition: ResponseHeaderBag::DISPOSITION_INLINE);
    }

    #[Route('/file/validate/{id}', name: '_validate_file', methods: ['PUT'])]
    public function verifyFile(File $file, EntityManagerInterface $manager): Response
    {
        $file->validate();
        $manager->persist($file);
        $manager->flush();
        return new Response();
    }

    #[Route('/{id}/edit', name: '_edit', methods: ['GET', 'POST'])]
    public function editApplication(ParkingApplication $parkingApp, Request $request, ApplicationService $applicationService, TranslatorInterface $translator)
    {
        $form = $this->createForm(ParkingApplicationType::class, $parkingApp);

        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid() && $applicationService->processForm($form, $form->getData(), $request, true))
        {
            $this->addFlash("success", $translator->trans("flash.application-modified"));
            return $this->redirectToRoute('admin_application_get', ['id' => $parkingApp->getId()]);
        }
        else
        {
            return $this->renderForm('application/index.html.twig', ['form' => $form]);
        }
    }
}
