<?php

namespace App\Controller\Admin;

use App\Entity\HomeMessage;
use App\Entity\Session;
use App\Form\HomeMessageType;
use App\Form\SessionType;
use App\Repository\HomeMessageRepository;
use App\Repository\SessionRepository;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\Exception\BadRequestException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

#[Route('/admin/config', name: 'admin_config_')]
class ConfigController extends AbstractController
{
    #[Route('/session', name: 'session', methods: ['GET'])]
    public function index(SessionRepository $sessionRepository): Response
    {
        $sessions = $sessionRepository->findAllOrderedByRecent();
        $activeSession = $sessionRepository->findActive();

        $form = $this->createForm(SessionType::class);

        return $this->render('admin/config/session.html.twig', [
            'form' => $form->createView(),
            'sessions' => $sessions,
            'activeSession' => $activeSession
        ]);
    }

    #[Route('/session', name: 'session_create', methods: ['POST'])]
    public function sessionCreate(
        Request $request,
        EntityManagerInterface $manager,
        SessionRepository $sessionRepository
    ): Response {

        $activeSession = $sessionRepository->findActive();
        if ($activeSession != null) {
            return new JsonResponse(["status" => "failure", "message" => "active.session.exists"], 400);
        }

        $session = new Session();

        $form = $this->createForm(SessionType::class, $session);
        $form->handleRequest($request);

        if ($form->isSubmitted()) {
            if (!$form->isValid()) {
                return new JsonResponse(["status" => "failure", "message" => "form.invalid"], 400);
            }

            $manager->persist($session);
            $manager->flush();
        }

        return new JsonResponse(["status" => "success"]);
    }

    #[Route('/session/archive', name: 'session_archive', methods: ['PUT'])]
    public function sessionArchive(SessionRepository $sessionRepository, EntityManagerInterface $manager): Response
    {
        $session = $sessionRepository->findActive();

        if ($session == null) {
            throw $this->createNotFoundException("No unarchived session");
        }

        $session->setArchived(true);

        $manager->persist($session);
        $manager->flush();

        return new JsonResponse(["status" => "success"]);
    }

    #[Route('/session/unarchive/{id}', name: 'session_unarchive', methods: ['PUT'])]
    public function sessionUnarchive(Session $session, SessionRepository $sessionRepository, EntityManagerInterface $manager)
    {
        $activeSession = $sessionRepository->findActive();

        if ($activeSession != null) {
            throw new BadRequestException("There already is an active session !");
        }

        $session->setArchived(false);

        $manager->persist($session);
        $manager->flush();

        return new JsonResponse(["status" => "success"]);
    }

    #[Route('/site', name: 'site')]
    public function site(Request $request, EntityManagerInterface $manager, HomeMessageRepository $homeMessageRepository): Response
    {
        $homeMessages = $homeMessageRepository->findAll();

        if (!$homeMessages) {
            $homeMessage = new HomeMessage();
        } elseif (count($homeMessages) == 1) {
            $homeMessage = $homeMessages[0];
        } else {
            foreach ($homeMessages as $message) {
                $manager->delete($message);
            }
            $homeMessage = new HomeMessage();
        }

        $form = $this->createForm(HomeMessageType::class, $homeMessage);
        $form->handleRequest($request);

        if ($form->isSubmitted() && $form->isValid()) {
            $manager->persist($homeMessage);
            $manager->flush();
            
            $this->addFlash('success', 'Le message d\'accueil a été modifé');
            return $this->redirectToRoute("admin_config_site");
        }

        return $this->render('admin/config/site.html.twig', [
            "form" => $form->createView()
        ]);
    }
}
