<?php

namespace App\Entity;

use App\Enum\ApplicationStatusEnum;
use App\Enum\AssociationEnum;
use App\Repository\ParkingApplicationRepository;
use Doctrine\ORM\Mapping as ORM;
use Symfony\Component\Validator\Constraints as Assert;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\Common\Collections\Collection;
use Symfony\Bridge\Doctrine\Validator\Constraints\UniqueEntity;
use Symfony\Component\Security\Core\User\UserInterface;
use Symfony\Component\Validator\Context\ExecutionContextInterface;

#[ORM\Entity(repositoryClass: ParkingApplicationRepository::class)]
#[ORM\HasLifecycleCallbacks]
#[UniqueEntity(
    fields: ["email", "session"],
    errorPath: "email",
    message: "Une demande existe déjà pour cet adresse courriel"
)]
class ParkingApplication implements UserInterface
{
    public function __construct()
    {
        $this->setCreatedDate(new \DateTime());
        $this->setUpdatedDate(new \DateTime());
        $this->vehicles = new ArrayCollection();
        $this->addressProof = new ArrayCollection();
        $this->parentProof = new ArrayCollection();
        $this->physicalConditionProof = new ArrayCollection();
        $this->emailAccessTokens = new ArrayCollection();
    }

    #[Assert\Callback]
    public function validate(ExecutionContextInterface $context, $payload)
    {
        if ($this->carpooling && ($this->carpoolDescription == null || strlen($this->carpoolDescription) == 0))
            $context->buildViolation("parking-app.carpool.description.notblank")
                ->atPath('carpoolDescription')
                ->addViolation();

        if ($this->parent && ($this->parentProof == null || count($this->parentProof) == 0))
            $context->buildViolation("parking-app.parent.proof.notblank")
                ->atPath('parentProof')
                ->addViolation();

        if ($this->physicalCondition && ($this->physicalConditionProof == null || count($this->physicalConditionProof) == 0))
            $context->buildViolation("parking-app.physical-condition.proof.notblank")
                ->atPath('physicalConditionProof')
                ->addViolation();

        // if (strpos($this->email, '@') === false || explode('@', $this->email)[1] != "polymtl.ca")
        //     $context->buildViolation("parking-app.email.polymtl")
        //         ->atPath('email')
        //         ->addViolation();
    }

    /**
     * @see UserInterface
     */
    public function getRoles(): array
    {
        return ["ROLE_USER"];
    }

    /**
     * @see UserInterface
     */
    public function getUserIdentifier(): string
    {
        return $this->email;
    }

    /**
     * @see UserInterface
     */
    public function getUsername(): string
    {
        return $this->getUserIdentifier();
    }

    /**
     * @see UserInterface
     */
    public function getPassword(): ?string
    {
        return null;
    }

    /**
     * @see UserInterface
     */
    public function getSalt(): ?string
    {
        return null;
    }

    /**
     * @see UserInterface
     */
    public function eraseCredentials()
    {
    }

    #[ORM\PrePersist]
    #[ORM\PreUpdate]
    public function updateModifiedTime()
    {
        $this->setUpdatedDate(new \DateTime());
    }

    #[ORM\Id]
    #[ORM\GeneratedValue]
    #[ORM\Column(type: "integer")]
    private $id;

    #[ORM\Column(type: "string", length: 255)]
    #[Assert\NotBlank]
    #[Assert\Length(max: 255)]
    private $forename;

    #[ORM\Column(type: "string", length: 255)]
    #[Assert\NotBlank()]
    #[Assert\Length(max: 255)]
    private $surname;

    #[ORM\Column(type: "string", length: 255)]
    #[Assert\NotBlank()]
    #[Assert\Email()]
    // Min email length is 14 assuming valid polymtl email such as: x.x@polymtl.ca
    #[Assert\Length(min: 14, max: 255)]
    private $email;

    #[ORM\Column(type: "string", length: 255)]
    #[Assert\NotBlank()]
    #[Assert\Length(min: 7, max: 7)]
    private $studentId;

    #[ORM\Column(type: "string", length: 255)]
    #[Assert\NotBlank()]
    #[Assert\Length(min: 3)]
    private $phone;

    #[ORM\OneToMany(targetEntity: Vehicle::class, mappedBy: "application", orphanRemoval: true, cascade: ["persist"])]
    #[Assert\Valid]
    #[Assert\Count(
        min: 1,
        minMessage: "parking-app.vehicle.min1",
        max: 10,
        maxMessage: "parking-app.vehicle.max10"
    )]
    private $vehicles;

    #[ORM\Column(type: "boolean")]
    #[Assert\NotNull()]
    private $carpooling;

    #[ORM\Column(type: "string", length: 255, nullable: true)]
    private $carpoolDescription;

    #[ORM\Column(type: "boolean")]
    #[Assert\NotNull()]
    private $parent;

    #[ORM\Column]
    #[Assert\NotNull()]
    private AssociationEnum $association;

    #[ORM\Column(type: "datetime")]
    #[Assert\NotBlank()]
    private $createdDate;

    #[ORM\Column(type: "datetime")]
    #[Assert\NotBlank()]
    private $updatedDate;

    #[ORM\Column(type: "boolean")]
    #[Assert\NotNull()]
    private $physicalCondition;

    #[ORM\OneToMany(targetEntity: AddressFile::class, mappedBy: "application", orphanRemoval: true, cascade: ["persist"])]
    #[Assert\NotBlank()]
    #[Assert\Count(
        min: 1,
        minMessage: "parking-app.address.proof.min1",
        max: 10,
        maxMessage: "parking-app.address.proof.max10"
    )]
    private $addressProof;

    #[ORM\OneToMany(targetEntity: ParentFile::class, mappedBy: "application", orphanRemoval: true, cascade: ["persist"])]
    #[Assert\Count(max: 10)]
    private $parentProof;

    #[ORM\OneToMany(
        targetEntity: PhysicalConditionFile::class,
        mappedBy: "application",
        orphanRemoval: true,
        cascade: ["persist"]
    )]
    #[Assert\Count(max: 10)]
    private $physicalConditionProof;

    #[ORM\Column]
    #[Assert\NotNull()]
    private ApplicationStatusEnum $status = ApplicationStatusEnum::CREATED;

    #[ORM\OneToMany(
        targetEntity: EmailAccessToken::class,
        mappedBy: "application",
        orphanRemoval: true,
        cascade: ["persist"]
    )]
    private $emailAccessTokens;

    #[ORM\Column(type: "text", nullable: true)]
    private $address;

    #[ORM\Column(type: "string", length: 255, nullable: true)]
    #[Assert\NotBlank()]
    private $street;

    #[ORM\Column(type: "string", length: 255, nullable: true)]
    #[Assert\NotBlank()]
    private $city;

    #[ORM\Column(type: "string", length: 10, nullable: true)]
    #[Assert\NotBlank()]
    #[Assert\Length(min: 5, max: 7)]
    private $postalCode;

    #[ORM\Column(type: "string", length: 255, nullable: true)]
    #[Assert\NotBlank()]
    private $province;

    #[ORM\Column(type: "string", length: 255, nullable: true)]
    #[Assert\NotBlank()]
    private $country;

    #[ORM\ManyToOne(targetEntity: Session::class, inversedBy: 'parkingApplications')]
    #[ORM\JoinColumn(nullable: false)]
    #[Assert\NotNull()]
    private $session;

    #[ORM\OneToOne(targetEntity: TravelInfo::class, cascade: ['persist', 'remove'])]
    #[ORM\JoinColumn(nullable: false)]
    private $travelInfo;

    public function getId(): ?int
    {
        return $this->id;
    }

    public function getForename(): ?string
    {
        return $this->forename;
    }

    public function setForename(string $forename): self
    {
        $this->forename = $forename;

        return $this;
    }

    public function getSurname(): ?string
    {
        return $this->surname;
    }

    public function setSurname(string $surname): self
    {
        $this->surname = $surname;

        return $this;
    }

    public function getEmail(): ?string
    {
        return $this->email;
    }

    public function setEmail(string $email): self
    {
        $this->email = $email;

        return $this;
    }

    public function getStudentId(): ?string
    {
        return $this->studentId;
    }

    public function setStudentId(string $studentId): self
    {
        $this->studentId = $studentId;

        return $this;
    }

    public function getPhone(): ?string
    {
        return $this->phone;
    }

    public function setPhone(string $phone): self
    {
        $this->phone = $phone;

        return $this;
    }

    /**
     * @return Collection|Vehicle[]
     */
    public function getVehicles(): Collection
    {
        return $this->vehicles;
    }

    public function addVehicle(Vehicle $vehicle): self
    {
        if (!$this->vehicles->contains($vehicle)) {
            $this->vehicles[] = $vehicle;
            $vehicle->setApplication($this);
        }

        return $this;
    }

    public function removeVehicle(Vehicle $vehicle): self
    {
        if ($this->vehicles->removeElement($vehicle)) {
            // set the owning side to null (unless already changed)
            if ($vehicle->getApplication() === $this) {
                $vehicle->setApplication(null);
            }
        }

        return $this;
    }

    public function getCarpooling(): ?bool
    {
        return $this->carpooling;
    }

    public function setCarpooling(bool $carpooling): self
    {
        $this->carpooling = $carpooling;

        return $this;
    }

    public function getCarpoolDescription(): ?string
    {
        return $this->carpoolDescription;
    }

    public function setCarpoolDescription(?string $carpoolDescription): self
    {
        $this->carpoolDescription = $carpoolDescription;

        return $this;
    }

    public function getParent(): ?bool
    {
        return $this->parent;
    }

    public function setParent(bool $parent): self
    {
        $this->parent = $parent;

        return $this;
    }

    /**
     * @return Collection|ParentFile[]
     */
    public function getParentProof(): Collection
    {
        return $this->parentProof;
    }

    public function addParentProof(File $parentProof): self
    {
        $parentProof = new ParentFile($parentProof);
        if (!$this->parentProof->contains($parentProof)) {
            $this->parentProof[] = $parentProof;
            $parentProof->setApplication($this);
        }

        return $this;
    }

    public function removeParentProof(ParentFile $parentProof): self
    {
        if ($this->parentProof->removeElement($parentProof)) {
            // set the owning side to null (unless already changed)
            if ($parentProof->getApplication() === $this) {
                $parentProof->setApplication(null);
            }
        }
        return $this;
    }

    public function getAssociation(): ?AssociationEnum
    {
        return $this->association;
    }

    public function setAssociation(AssociationEnum $association): self
    {
        $this->association = $association;

        return $this;
    }

    public function getCreatedDate(): ?\DateTimeInterface
    {
        return $this->createdDate;
    }

    public function setCreatedDate(\DateTimeInterface $createdDate): self
    {
        $this->createdDate = $createdDate;

        return $this;
    }

    public function getUpdatedDate(): ?\DateTimeInterface
    {
        return $this->updatedDate;
    }

    public function setUpdatedDate(\DateTimeInterface $updatedDate): self
    {
        $this->updatedDate = $updatedDate;

        return $this;
    }

    public function getPhysicalCondition(): ?bool
    {
        return $this->physicalCondition;
    }

    public function setPhysicalCondition(bool $physicalCondition): self
    {
        $this->physicalCondition = $physicalCondition;

        return $this;
    }

    /**
     * @return Collection|PhysicalConditionFile[]
     */
    public function getPhysicalConditionProof(): Collection
    {
        return $this->physicalConditionProof;
    }

    public function addPhysicalConditionProof(File $physicalConditionProof): self
    {
        $physicalConditionProof = new PhysicalConditionFile($physicalConditionProof);
        if (!$this->physicalConditionProof->contains($physicalConditionProof)) {
            $this->physicalConditionProof[] = $physicalConditionProof;
            $physicalConditionProof->setApplication($this);
        }

        return $this;
    }

    public function removePhysicalConditionProof(PhysicalConditionFile $physicalConditionProof): self
    {
        if ($this->physicalConditionProof->removeElement($physicalConditionProof)) {
            // set the owning side to null (unless already changed)
            if ($physicalConditionProof->getApplication() === $this) {
                $physicalConditionProof->setApplication(null);
            }
        }
        return $this;
    }

    /**
     * @return Collection|AddressFile[]
     */
    public function getAddressProof(): Collection
    {
        return $this->addressProof;
    }

    public function addAddressProof(File $addressProof): self
    {
        $addressProof = new AddressFile($addressProof);
        if (!$this->addressProof->contains($addressProof)) {
            $this->addressProof[] = $addressProof;
            $addressProof->setApplication($this);
        }

        return $this;
    }

    public function removeAddressProof(AddressFile $addressProof): self
    {
        if ($this->addressProof->removeElement($addressProof)) {
            // set the owning side to null (unless already changed)
            if ($addressProof->getApplication() === $this) {
                $addressProof->setApplication(null);
            }
        }

        return $this;
    }

    /**
     * @return Collection|EmailAccessToken[]
     */
    public function getEmailAccessTokens(): Collection
    {
        return $this->emailAccessTokens;
    }

    public function addEmailAccessToken(EmailAccessToken $emailAccessToken): self
    {
        if (!$this->emailAccessTokens->contains($emailAccessToken)) {
            $this->emailAccessTokens[] = $emailAccessToken;
            $emailAccessToken->setApplication($this);
        }
        return $this;
    }

    public function removeEmailAccessToken(EmailAccessToken $emailAccessToken): self
    {
        if ($this->emailAccessTokens->removeElement($emailAccessToken)) {
            // set the owning side to null (unless already changed)
            if ($emailAccessToken->getApplication() === $this) {
                $emailAccessToken->setApplication(null);
            }
        }
        return $this;
    }

    public function getStatus(): ?ApplicationStatusEnum
    {
        return $this->status;
    }

    public function setStatus(ApplicationStatusEnum $status): self
    {
        $this->status = $status;
        return $this;
    }

    public function getAddress(): ?string
    {
        if (!$this->street)
        {
            return $this->address;
        }
        else
        {
            return implode(", ", [
                $this->street,
                $this->city,
                $this->province,
                $this->postalCode,
                $this->country
            ]);
        }
    }

    public function splitAddress()
    {
        if (!$this->address || $this->country || $this->postalCode || $this->province || $this->city || $this->street) return;
        $values = explode(", ", $this->address);
        $this->country = array_pop($values);
        $this->postalCode = array_pop($values);
        $this->province = array_pop($values);
        $this->city = array_pop($values);
        $this->street = implode(", ", $values);
    }

    public function getStreet(): ?string
    {
        return $this->street;
    }

    public function setStreet(string $street): self
    {
        $this->street = $street;
        return $this;
    }

    public function getCity(): ?string
    {
        return $this->city;
    }

    public function setCity(string $city): self
    {
        $this->city = $city;
        return $this;
    }

    public function getProvince(): ?string
    {
        return $this->province;
    }

    public function setProvince(string $province): self
    {
        $this->province = $province;
        return $this;
    }

    public function getPostalCode(): ?string
    {
        return $this->postalCode;
    }

    public function setPostalCode(string $postalCode): self
    {
        $this->postalCode = $postalCode;
        return $this;
    }

    public function getCountry(): ?string
    {
        return $this->country;
    }

    public function setCountry(string $country): self
    {
        $this->country = $country;
        return $this;
    }

    static function sigmoid(float $x): float
    {
        $ex = \exp($x);
        return 2 * ($ex / (1 + $ex) - 0.5);
    }

    public function getScore(): float
    {
        // A lot of hyper parameters will be used in this function, these can be optimized once we have enough data.
        // Apply logistic sigmoid function to each of the values (to clamp them between 0 and 1)

        $DIST_DIV = 20.; // we want 40 Km to be close to 1
        $DUR_DRIVE_DIV = 60. * 30.; // we want 1 H to be close to 1
        $DUR_DRIVE_TRAF_DIV = 60. * 30.; // we want 1 H to be close to 1
        $DUR_TRANS_DIV = 60. * 30.; // we want 1 H to be close to 1

        $sigDist = $this->sigmoid($this->getTravelInfo()->getDistance() / $DIST_DIV);
        $sigDurationDrive = $this->sigmoid($this->getTravelInfo()->getDurationDriving() / $DUR_DRIVE_DIV);
        $sigDurationDriveTraffic = $this->sigmoid($this->getTravelInfo()->getDurationDrivingTraffic() / $DUR_DRIVE_TRAF_DIV);

        $durationTransit = $this->getTravelInfo()->getDurationTransit();
        $sigDurationTransit = ($durationTransit == -1) ? 1 : $this->sigmoid($durationTransit / $DUR_TRANS_DIV);


        // This defines the order of importance of each parameter.
        /// Once we have enough data, these parameters can be learned with a simple Feed forward Network
        $DIST_COEF = .2;
        $DUR_DRIVE_COEF = .1;
        $DUR_DRIVE_TRAF_COEF = .35;
        $DUR_TRANS_COEF = .35;

        return ($sigDist * $DIST_COEF +
            $sigDurationDrive * $DUR_DRIVE_COEF +
            $sigDurationDriveTraffic * $DUR_DRIVE_TRAF_COEF +
            $sigDurationTransit * $DUR_TRANS_COEF) * 100;
    }

    public function getSession(): ?Session
    {
        return $this->session;
    }

    public function setSession(?Session $session): self
    {
        $this->session = $session;

        return $this;
    }

    public function getTravelInfo(): ?TravelInfo
    {
        return $this->travelInfo;
    }

    public function setTravelInfo(TravelInfo $travelInfo): self
    {
        $this->travelInfo = $travelInfo;

        return $this;
    }
}
