<?php

namespace App\Repository;

use App\Entity\ParkingApplication;
use App\Entity\Session;
use App\Enum\ApplicationStatusEnum;
use Doctrine\Bundle\DoctrineBundle\Repository\ServiceEntityRepository;
use Doctrine\Persistence\ManagerRegistry;
use Doctrine\Common\Collections\Criteria;

/**
 * @method ParkingApplication|null find($id, $lockMode = null, $lockVersion = null)
 * @method ParkingApplication|null findOneBy(array $criteria, array $orderBy = null)
 * @method ParkingApplication[]    findAll()
 * @method ParkingApplication[]    findBy(array $criteria, array $orderBy = null, $limit = null, $offset = null)
 */
class ParkingApplicationRepository extends ServiceEntityRepository
{
    public function __construct(ManagerRegistry $registry)
    {
        parent::__construct($registry, ParkingApplication::class);
    }

    /**
     * @return ParkingApplication[] Returns an array of ParkingApplication objects
     */
    public function findOrderedByTime()
    {
        return $this->createQueryBuilder('p')
            ->orderBy('p.travelTime', Criteria::DESC)
            ->getQuery()
            ->getResult();
    }

    /**
     * @return ParkingApplication[] Returns an array of ParkingApplication objects
     */
    public function findByStatus(ApplicationStatusEnum $status)
    {
        return $this->createQueryBuilder('p')
            ->join('p.session', 's')
            ->andWhere('p.status = :status')
            ->andWhere('s.archived = 0')
            ->setParameter('status', $status->value)
            ->getQuery()
            ->getResult();
    }

    /**
     * @return ParkingApplication[] Returns an array of ParkingApplication objects
     */
    public function findBySession(Session $session)
    {
        return $this->createQueryBuilder('p')
            ->join('p.session', 's')
            ->andWhere('s.id = :session')
            ->setParameter('session', $session->getId())
            ->getQuery()
            ->getResult();
    }

    /**
     * @return array<string, int> Map of status to the number of applications with that status
     */
    public function countByStatus(Session $session)
    {
        $counts = $this->createQueryBuilder('p')
            ->select('p.status, COUNT(p.status) as count')
            ->join('p.session', 's')
            ->andWhere('s.id = :session')
            ->groupBy('p.status')
            ->setParameter('session', $session->getId())
            ->getQuery()
            ->getResult();

        $counts_map = [];
        foreach ($counts as $count) {
            $counts_map[$count['status']->value] = $count['count'];
        }

        foreach (ApplicationStatusEnum::cases() as $status) {
            if (!array_key_exists($status->value, $counts_map)) {
                $counts_map[$status->value] = 0;
            }
        }

        return $counts_map;
    }

    /**
     * @return ParkingApplication[] Returns an array of ParkingApplication objects
     */
    public function findActiveByEmailOrderedByCreationDate(string $email)
    {
        return $this->createQueryBuilder('p')
            ->Where('p.email = :email')
            ->join('p.session', 's')
            ->andWhere('s.archived = 0')
            ->setParameter('email', $email)
            ->orderBy('p.createdDate', Criteria::DESC)
            ->getQuery()
            ->getResult();
    }
}
