<?php

namespace App\Security;

use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Symfony\Component\Security\Core\Exception\AuthenticationException;
use Symfony\Component\Security\Core\Exception\CustomUserMessageAuthenticationException;
use Symfony\Component\Security\Http\Authenticator\AbstractAuthenticator;
use Symfony\Component\Security\Http\Authenticator\Passport\PassportInterface;
use Symfony\Component\Security\Http\Authenticator\Passport\Badge\UserBadge;
use Symfony\Component\Security\Http\Authenticator\Passport\SelfValidatingPassport;
use App\Repository\EmailAccessTokenRepository;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Session\Session;
use Symfony\Component\Routing\RouterInterface;
use Symfony\Component\Security\Http\EntryPoint\AuthenticationEntryPointInterface;
use Symfony\Contracts\Translation\TranslatorInterface;

class UserLinkAuthenticator extends AbstractAuthenticator implements AuthenticationEntryPointInterface
{
    private EmailAccessTokenRepository $emailAccessTokenRepository;
    private RouterInterface $router;
    private TranslatorInterface $translator;


    public function __construct(EmailAccessTokenRepository $emailAccessTokenRepository, RouterInterface $router, TranslatorInterface $translator)
    {
        $this->emailAccessTokenRepository = $emailAccessTokenRepository;
        $this->router = $router;
        $this->translator = $translator;
    }

    private function getSession($request): Session
    {
        $session = $request->getSession();
        if (!$session instanceof Session) {
            throw new \Exception("SessionInterface must be Symfony Session");
        }
        return $session;
    }

    public function supports(Request $request): ?bool
    {
        return $request->query->has("token");
    }

    public function authenticate(Request $request): PassportInterface
    {
        $token = $request->query->get('token');

        if ($token === NULL) {
            throw new CustomUserMessageAuthenticationException("No token provided");
        }

        return new SelfValidatingPassport(
            new UserBadge($token, function ($userToken) {
                if (($emailToken = $this->emailAccessTokenRepository->findOneByTokenNotExpired($userToken)) !== null)
                    return $emailToken->getApplication();
            })
        );
    }

    public function onAuthenticationSuccess(Request $request, TokenInterface $token, string $firewallName): ?Response
    {
        return null;
    }

    public function onAuthenticationFailure(Request $request, AuthenticationException $exception): ?Response
    {
        $this->getSession($request)->getFlashBag()->add(
           'error',
           $this->translator->trans("flash.auth-failed")
        );
        return new RedirectResponse($this->router->generate("application_login"));
    }

    public function start(Request $request, ?AuthenticationException $authException = null): ?Response
    {
        $url = $request->attributes->get("_route");

        return new RedirectResponse($this->router->generate("application_login", ["redirection_url" => $url ]));
    }
}
