<?php

namespace App\Service;

use App\Entity\ParkingApplication;

class LocalisationService
{
    private string $bingMapsApiKey;
    private string $polyCoords;
    private string $endpointUrl;
    private string $routeTransitEndpoint;
    private string $routeDrivingEndpoint;
    private string $locationEndpoint;

    public function __construct(
        string $bingMapsApiKey,
        string $polyCoords,
        string $endpointUrl,
        string $routeTransitEndpoint,
        string $routeDrivingEndpoint,
        string $locationEndpoint
    ) {
        $this->bingMapsApiKey = $bingMapsApiKey;
        $this->polyCoords = $polyCoords;
        $this->endpointUrl = $endpointUrl;
        $this->routeTransitEndpoint = $routeTransitEndpoint;
        $this->routeDrivingEndpoint = $routeDrivingEndpoint;
        $this->locationEndpoint = $locationEndpoint;
    }

    private function getMondayBeforeSessionEnd(ParkingApplication $parkingApp): ?string
    {
        $session = $parkingApp->getSession();
        if ($session === null) {
            return null;
        }

        $endDate = $session->getEndDate();
        if ($endDate === null) {
            return null;
        }

        return date('m/d/Y', strtotime('previous monday', $endDate->getTimestamp()));
    }

    private function callCurl(string $url): ?object
    {
        $curl = curl_init($url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);

        $resp = curl_exec($curl);
        $response = "";

        if (!curl_errno($curl)) {
            switch ($errorCode = curl_getinfo($curl, CURLINFO_HTTP_CODE)) {
                case 200:
                    $response = json_decode($resp);
                    break;
                default:
                    \Sentry\captureMessage("Got code " . $errorCode . " when curling " . $url);
                    curl_close($curl);
                    return null;
            }
        } else {
            \Sentry\captureMessage("Got error " . curl_error($curl) . " when curling " . $url);
            curl_close($curl);
            return null;
        }

        curl_close($curl);

        return $response;
    }

    //TODO Find a free address validator API if we want more precision in the future.

    public function getTravelInformation(ParkingApplication $parkingApp): array
    {
        $coords = $this->getAdressGeolocalisation($parkingApp->getAddress());

        if (!$coords) {
            return [];
        }

        $mondayBeforeSessionEnd = $this->getMondayBeforeSessionEnd($parkingApp);
        if ($mondayBeforeSessionEnd === null) {
            return [];
        }

        // Documentation : https://docs.microsoft.com/en-us/bingmaps/rest-services/routes/calculate-a-route
        $body = [
            'wayPoint.1' => $coords,
            'waypoint.2' => $this->polyCoords,
            'routeAttributes' => 'routePath',
            'timeType' => 'Arrival',
            'dateTime' => $mondayBeforeSessionEnd . ' 08:30:00',
            'maxSolutions' => '1',
            'key' => $this->bingMapsApiKey
        ];

        $drivingUrl = $this->endpointUrl . $this->routeDrivingEndpoint . http_build_query($body);
        $transitUrl = $this->endpointUrl . $this->routeTransitEndpoint . http_build_query($body);

        $drivingResult = $this->callCurl($drivingUrl);
        $transitResult = $this->callCurl($transitUrl);

        if (!$drivingResult) {
            return [];
        }

        $distance = $drivingResult->resourceSets[0]->resources[0]->travelDistance;
        $durationDriving = $drivingResult->resourceSets[0]->resources[0]->travelDuration;
        $durationDrivingTraffic = $drivingResult->resourceSets[0]->resources[0]->travelDurationTraffic;
        $routePoints = $drivingResult->resourceSets[0]->resources[0]->routePath->line->coordinates;

        $durationTransit = -1;
        if ($transitResult !== null) {
            $durationTransit = $transitResult->resourceSets[0]->resources[0]->travelDuration;
        }

        return [
            "distance" => $distance,
            "durationDriving" => $durationDriving,
            "durationDrivingTraffic" => $durationDrivingTraffic,
            "routePoints" => $routePoints,
            "durationTransit" => $durationTransit
        ];
    }

    private function getAdressGeolocalisation(string $address): ?string
    {
        // Documentation : https://docs.microsoft.com/en-us/bingmaps/rest-services/locations/
        $body = [
            'q' => $address,
            'key' => $this->bingMapsApiKey,
        ];

        $url = $this->endpointUrl . $this->locationEndpoint . http_build_query($body);

        $response = $this->callCurl($url);

        if ($response === null || !count($response->resourceSets) || !count($response->resourceSets[0]->resources)) {
            return null;
        }

        $latitude = $response->resourceSets[0]->resources[0]->point->coordinates[0];
        $longitude = $response->resourceSets[0]->resources[0]->point->coordinates[1];

        $coordinates = strval($latitude) . ',' . strval($longitude);
        return $coordinates;
    }
}
